# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
"""
This is then main function for open any files in local or remote space
with the best python libraries and the best practice such as build-in
io.open, mmap, etc.
"""
import abc
import csv
import io
import json
import marshal
import mmap
import os
import pickle
from contextlib import contextmanager
from pathlib import Path
from types import ModuleType
from typing import (
    IO,
    Any,
    AnyStr,
    Callable,
    Dict,
    List,
    Literal,
    NoReturn,
    Optional,
    Union,
    get_args,
)

# TODO: import msgpack
import yaml

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader

from .settings import SettingRegex
from .utils import search_env, search_env_replace

FileCompressType = Literal["gzip", "gz", "xz", "bz2"]

# TODO: add more compress type such as
#  - h5,hdf5(h5py)
#  - fits(astropy)
#  - rar(...)
DirCompressType = Literal["zip", "rar", "tar"]


class CompressModule(ModuleType):
    def decompress(self, *args, **kwargs) -> AnyStr:
        ...

    def open(self, *args, **kwargs) -> IO:
        ...


class BaseFile(abc.ABC):
    @abc.abstractmethod
    def read(self, *args, **kwargs):
        ...

    @abc.abstractmethod
    def write(self, *args, **kwargs):
        ...


class OpenFile(BaseFile):
    """Open File Object"""

    def __init__(
        self,
        path: Union[str, Path],
        *,
        encoding: Optional[str] = None,
        compress: Optional[FileCompressType] = None,
    ):
        self.path = path
        self.encoding = encoding or "utf-8"
        self.compress = compress

        # Action anything after set up attributes.
        self._do_after_set_attrs()

    def _do_after_set_attrs(self) -> NoReturn:
        ...

    @property
    def compress_lib(self) -> CompressModule:
        """Return Compress package"""
        if not self.compress:
            return io
        elif self.compress in {"gzip", "gz"}:
            import gzip

            return gzip
        elif self.compress in {"bz2"}:
            import bz2

            return bz2
        elif self.compress in {"xz"}:
            import lzma as xz

            return xz
        raise NotImplementedError

    @property
    def decompress(self) -> Callable:
        if self.compress and self.compress in get_args(FileCompressType):
            return self.compress_lib.decompress
        raise NotImplementedError(
            "Does not implement decompress method for None compress value."
        )

    def convert_mode(
        self,
        mode: Optional[str] = None,
        default: bool = True,
    ) -> Dict[str, str]:
        if not mode:
            if default:
                return {"mode": "r"}
            raise ValueError("The mode value does not set.")
        byte_mode: bool = "b" in mode
        if self.compress is None:
            _mode: Dict[str, str] = {"mode": mode}
            return _mode if byte_mode else {"encoding": self.encoding, **_mode}
        elif not byte_mode and self.compress in {
            "gzip",
            "gz",
            "xz",
            "bz2",
        }:
            # Add `t` in open file mode for force with text mode.
            return {"mode": f"{mode}t", "encoding": self.encoding}
        elif byte_mode and self.compress in {
            "gzip",
            "gz",
            "xz",
            "bz2",
        }:
            return {"mode": mode}

    def open(self, *, mode: Optional[str] = None, **kwargs) -> IO:
        return self.compress_lib.open(
            self.path,
            **self.convert_mode(mode),
            **kwargs,
        )

    @contextmanager
    def mem_open(self, *, mode: Optional[str] = None) -> IO:
        _mode: str = mode or "r"
        _f: IO = self.open(mode=mode)
        _access = mmap.ACCESS_READ if ("r" in _mode) else mmap.ACCESS_WRITE
        try:
            yield mmap.mmap(
                _f.fileno(),
                length=0,
                access=_access,
            )
        except ValueError:
            yield _f
        finally:
            _f.close()

    # def read(self, memory: bool = False):
    #     if not self._f:
    #         self._f = self.open()
    #     if not self.compress:
    #         return (
    #             self._f.read().decode(self.encoding)
    #             if memory
    #             else self._f.read()
    #         )
    #     elif self.compress in {
    #         "gzip",
    #         "gz",
    #         "xz",
    #         "bz2",
    #     }:
    #         if memory:
    #             return self.compress_lib.decompress(self._f.read()).decode(
    #                 self.encoding
    #             )
    #         return self._f.read()

    def read(self, *args, **kwargs):
        raise NotImplementedError

    def write(self, *args, **kwargs):
        raise NotImplementedError


class OpenDir:
    """Open File Object"""

    def __init__(
        self,
        path: Union[str, Path],
        *,
        compress: Optional[DirCompressType] = None,
    ):
        self.path = path
        self.compress = compress
        # Action anything after set up attributes.
        self._do_after_set_attrs()

    def _do_after_set_attrs(self) -> NoReturn:
        ...

    def open(self, *, mode: str, **kwargs):
        if not self.compress:
            return ""
        elif self.compress in {"zip"}:
            import zipfile

            return zipfile.ZipFile(
                self.path,
                mode=mode,
                compression=zipfile.ZIP_DEFLATED,
                **kwargs,
            )
        elif self.compress in {"tar"}:
            import tarfile

            # TODO: Wrapped tar module with change some methods like,
            #   - add --> write
            return tarfile.open(
                self.path,
                mode=f"{mode}:gz",  # w:bz2
            )
        return NotImplementedError


class Env(OpenFile):
    """Env object which mapping search engine"""

    KEEP_NEWLINE: bool = False
    DEFAULT: str = ""

    def read(self, *, update: bool = True) -> dict:
        with self.open(mode="r") as _r:
            print(_r.read())
            _r.seek(0)

            _result: dict = search_env(
                _r.read(),
                keep_newline=self.KEEP_NEWLINE,
                default=self.DEFAULT,
            )
            if update:
                os.environ.update(**_result)
            return _result

    def write(self, data: Dict[str, Any]) -> NoReturn:
        raise NotImplementedError


class Yaml(OpenFile):
    """Yaml File Object

    .. noted::
        - The boolean value in the yaml file
            - true: Y, true, Yes, ON
            - false: n, false, No, off
    """

    def read(
        self,
        safe: bool = True,
    ) -> Dict[str, Any]:
        if safe:
            with self.open(mode="r") as _r:
                return yaml.load(_r.read(), SafeLoader)
        return NotImplementedError

    def write(self, data: Dict[str, Any]) -> NoReturn:
        with self.open(mode="w") as _w:
            yaml.dump(data, _w, default_flow_style=False)


class YamlEnv(Yaml):
    """Yaml object which mapping search environment variable."""

    raise_if_not_default: bool = False
    default: str = "N/A"
    escape: str = "ESC"

    @staticmethod
    def prepare(x: str) -> str:
        return x

    def read(
        self,
        safe: bool = True,
    ) -> Dict[str, Any]:
        if safe:
            with self.open(mode="r") as _r:
                _env_replace: str = search_env_replace(
                    SettingRegex.RE_YAML_COMMENT.sub("", _r.read()),
                    raise_if_default_not_exists=self.raise_if_not_default,
                    default_value=self.default,
                    escape_replaced=self.escape,
                    caller=self.prepare,
                )
                if _result := yaml.load(_env_replace, SafeLoader):
                    return _result
                return {}
        return NotImplementedError

    def write(self, data: Dict[str, Any]) -> NoReturn:
        raise NotImplementedError


class CSV(OpenFile):
    def read(self) -> List[str]:
        with self.open(mode="r") as _r:
            try:
                dialect = csv.Sniffer().sniff(_r.read(128))
                _r.seek(0)
                return list(csv.DictReader(_r, dialect=dialect))
            except csv.Error:
                return []

    def write(
        self,
        data: Union[List[Any], Dict[Any, Any]],
        *,
        mode: Optional[str] = None,
        **kwargs,
    ) -> NoReturn:
        mode = mode or "w"
        assert mode in {
            "a",
            "w",
        }, "save mode must contain only value `a` nor `w`."
        with self.open(mode=mode, newline="") as _w:
            _has_data: bool = True
            if isinstance(data, dict):
                data: list = [data]
            elif not data:
                data: list = [{}]
                _has_data: bool = False
            if _has_data:
                writer = csv.DictWriter(
                    _w,
                    fieldnames=list(data[0].keys()),
                    lineterminator="\n",
                    **kwargs,
                )
                if mode == "w" or not self.has_header:
                    writer.writeheader()
                writer.writerows(data)

    @property
    def has_header(self) -> bool:
        with self.open(mode="r") as _r:
            try:
                return csv.Sniffer().has_header(_r.read(128))
            except csv.Error:
                return False


class CSVPipeDim(CSV):
    def _do_after_set_attrs(self) -> NoReturn:
        csv.register_dialect(
            "pipe_delimiter", delimiter="|", quoting=csv.QUOTE_ALL
        )

    def read(self) -> list:
        with self.open(mode="r") as _r:
            try:
                return list(
                    csv.DictReader(_r, delimiter="|", quoting=csv.QUOTE_ALL)
                )
            except csv.Error:
                return []

    def write(
        self,
        data: Union[List[Any], Dict[Any, Any]],
        *,
        mode: Optional[str] = None,
        **kwargs,
    ) -> NoReturn:
        mode = mode or "w"
        assert mode in {
            "a",
            "w",
        }, "save mode must contain only value `a` nor `w`."
        with self.open(mode=mode, newline="") as _w:
            _has_data: bool = True
            if isinstance(data, dict):
                data: list = [data]
            elif not data:
                data: list = [{}]
                _has_data: bool = False
            if _has_data:
                writer = csv.DictWriter(
                    _w,
                    fieldnames=list(data[0].keys()),
                    lineterminator="\n",
                    delimiter="|",
                    quoting=csv.QUOTE_ALL,
                    **kwargs,
                )
                if mode == "w" or not self.has_header:
                    writer.writeheader()
                writer.writerows(data)


class Json(OpenFile):
    def read(self) -> Union[Dict[Any, Any], List[Any]]:
        with self.open(mode="r") as _r:
            try:
                return json.loads(_r.read())
            except json.decoder.JSONDecodeError:
                return {}

    def write(
        self,
        data,
        *,
        indent: int = 4,
    ) -> NoReturn:
        _w: IO
        with self.open(
            mode="w",
        ) as _w:
            if self.compress:
                _w.write(json.dumps(data))
            else:
                json.dump(data, _w, indent=indent)


class JsonEnv(Json):
    raise_if_not_default: bool = False
    default: str = "N/A"
    escape: str = "ESC"

    @staticmethod
    def prepare(x: str) -> str:
        return x

    def read(self) -> Union[Dict[Any, Any], List[Any]]:
        with self.open(mode="r") as _r:
            return json.loads(
                search_env_replace(
                    _r.read(),
                    raise_if_default_not_exists=self.raise_if_not_default,
                    default_value=self.default,
                    escape_replaced=self.escape,
                    caller=self.prepare,
                )
            )

    def write(self, data, *, indent: int = 4) -> NoReturn:
        raise NotImplementedError


class Pickle(OpenFile):
    def read(self):
        with self.open(mode="rb") as _r:
            return pickle.loads(_r.read())

    def write(self, data):
        with self.open(mode="wb") as _w:
            pickle.dump(data, _w)


class Marshal(OpenFile):
    def read(self):
        with self.open(mode="rb") as _r:
            return marshal.loads(_r.read())

    def write(self, data):
        with self.open(mode="wb") as _w:
            marshal.dump(data, _w)


__all__ = (
    "OpenFile",
    "Env",
    "Json",
    "JsonEnv",
    "Yaml",
    "YamlEnv",
    "CSV",
    "CSVPipeDim",
    "Marshal",
    "Pickle",
)



from .config import (
    ConfAdapter,
    ConfFile,
    ConfSQLite,
)
from .exceptions import (
    ConfigArgumentError,
    ConfigNotFound,
    IOBaseError,
)
from .models import (
    EngineData,
    FlagData,
    Params,
    PathData,
    RuleData,
    StageData,
    ValueData,
)
from .register import Register



import datetime
import fnmatch
import os
import re
import shutil
from pathlib import (
    Path,
    PosixPath,
)
from typing import (
    AnyStr,
    Callable,
    Iterator,
    Optional,
)

join_os: Callable = os.sep.join


def replace_os(path: str) -> str:
    return path.replace(os.sep, "/")


def join_path(
    full_path: AnyStr, full_join_path: str, abs: bool = True
) -> AnyStr:
    """Join path with multi pardir value if set `full_join_path`
    be '../../<path>'.
    """
    _abspath: AnyStr = full_path
    if re.search(r"^(\w+://)", _abspath) or (not abs):
        # Return joined value with '/' string value
        # if the full path starts with protocol prefix.
        return "/".join([replace_os(_abspath), replace_os(full_join_path)])
    _join_split: list = os.path.normpath(full_join_path).split(os.sep)
    for path in _join_split:
        _abspath: AnyStr = (
            os.path.abspath(os.path.join(_abspath, os.pardir))
            if path == ".."
            else os.path.abspath(os.path.join(_abspath, path))
        )
    return replace_os(_abspath)


def join_root_with(full_join_path: str, key_name: str = "APP_PATH") -> AnyStr:
    """Join path with the root path which set in `key_name` environment
    variable.
    """
    if key_value := os.getenv(key_name):
        return join_path(key_value, full_join_path)
    raise ValueError(
        f"the key name: {key_name!r} of root path in environment variable "
        f"does not exists"
    )


def get_modification_time(path: str):
    """Return datetime of modification of file."""
    timestamp = os.path.getmtime(path)
    return datetime.datetime.fromtimestamp(timestamp)


def get_files(path: str, pattern: str) -> Iterator[PosixPath]:
    """Return path from glob method."""
    yield from Path(path).glob(pattern)


def remove_file(path, is_dir: bool = False):
    """param <path> could either be relative or absolute."""
    if os.path.isfile(path) or os.path.islink(path):
        # remove the file
        os.remove(path)
    elif os.path.isdir(path) and is_dir:
        # remove dir and all contains
        shutil.rmtree(path)
    else:
        raise ValueError(
            f"file {path!r} is not a file{' or dir' if is_dir else ''}."
        )


def touch(filename: str, times=None):
    file_handle = open(filename, "a")
    try:
        os.utime(filename, times)
    finally:
        file_handle.close()


class PathSearch:
    """Path Search object"""

    @classmethod
    def from_dict(cls, _dict: dict):
        """Return Path Search object with dictionary"""
        return cls(
            root=_dict["root"],
            exclude_folder=_dict.get("exclude_folder"),
            exclude_name=_dict.get("exclude_name"),
            max_level=_dict.get("max_level", -1),
            length=_dict.get("length", 4),
            icon=_dict.get("icon", 1),
        )

    def __init__(
        self,
        root: str,
        *,
        exclude_name: Optional[list] = None,
        exclude_folder: Optional[list] = None,
        max_level: int = -1,
        length: int = 4,
        icon: int = 1,
    ):
        self.root: str = root
        self.exclude_folder: list = exclude_folder or []
        self.exclude_name: list = exclude_name or []
        self.max_level: int = max_level
        self.length: int = length
        self.real_level: int = 0
        self._icon_last: str = self.icons[icon]["last"]
        self._icon_next: str = self.icons[icon]["next"]
        self._icon: str = self.icons[icon]["normal"]
        self._icon_length: int = len(self._icon)
        assert (
            self._icon_length + 1
        ) < self.length, "a `length` argument must gather than length of icon."

        self.output_buf: list = [f"[{self.root.rsplit(os.path.sep, 1)[-1]}]"]
        self.output_files: list = []
        try:
            self._recurse(self.root, os.listdir(self.root), "", 0)
        except FileNotFoundError:
            pass

    @property
    def level(self) -> int:
        """Return level of sub path from the root path."""
        return self.real_level + 1 if self.max_level == -1 else self.max_level

    @property
    def files(self) -> list:
        """Return files which include in the root path."""
        return self.output_files

    @property
    def icons(self) -> dict:
        return {
            1: {"normal": "│", "next": "├─", "last": "└─"},
            2: {"normal": "┃", "next": "┣━", "last": "┗━"},
        }

    def _recurse(
        self, parent_path: str, file_list: list, prefix: str, level: int
    ):
        """Path recursive method for generate buffer of tree and files."""
        if not file_list or (self.max_level != -1 and self.max_level <= level):
            return

        self.real_level = max(level, self.real_level)
        file_list.sort(key=lambda f: os.path.isfile(join_path(parent_path, f)))
        for idx, sub_path in enumerate(file_list):
            if any(
                exclude_name in sub_path for exclude_name in self.exclude_name
            ):
                continue

            full_path: str = join_path(parent_path, sub_path)
            idc = self._switch_icon(idx, len(file_list))

            if os.path.isdir(full_path) and sub_path not in self.exclude_folder:
                self.output_buf.append(f"{prefix}{idc}[{sub_path}]")
                tmp_prefix: str = (
                    (
                        f"{prefix}{self._icon}"
                        f'{" " * (self.length - self._icon_length)}'
                    )
                    if len(file_list) > 1 and idx != len(file_list) - 1
                    else f'{prefix}{" " * self.length}'
                )
                self._recurse(
                    full_path, os.listdir(full_path), tmp_prefix, level + 1
                )
            elif os.path.isfile(full_path):
                self.output_buf.append(f"{prefix}{idc}{sub_path}")
                self.output_files.append(full_path)

    def pick(self, filename: str) -> list:
        """Return filename with match with input argument."""
        return list(
            filter(
                lambda file: fnmatch.fnmatch(file, f"*/{filename}"),
                self.output_files,
            )
        )

    def make_tree(self, newline: Optional[str] = None) -> str:
        """Return path tree of root path."""
        _newline: str = newline or "\n"
        return _newline.join(self.output_buf)

    def _switch_icon(self, number_now: int, number_all: int):
        return (
            self._icon_last
            if number_now == (number_all - 1)
            else self._icon_next
        )



from .files import (
    CSV,
    CSVPipeDim,
    Env,
    Json,
    JsonEnv,
    Marshal,
    OpenFile,
    Pickle,
    Yaml,
    YamlEnv,
)
from .pathutils import get_files
from .settings import SettingRegex
from .utils import (
    add_newline,
    search_env_replace,
)



import os.path
import pathlib
import shutil
import unittest

import ddeutil.io.config as conf
import pytest


@pytest.mark.usefixtures("test_path_to_cls")
class BaseConfigFileTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.demo_path: pathlib.Path = (
            self.test_path / "examples" / "conf" / "demo"
        )
        self.target_path: pathlib.Path = self.test_path / "conf_file_temp"

    def test_base_conf_read_file(self):
        bcf = conf.BaseConfFile(self.demo_path)

        self.assertDictEqual(
            {
                "alias": "conn_local_data_landing",
                "endpoint": "file:///N/A/data/demo/landing",
                "type": "connection.LocalSystem",
            },
            bcf.load(name="conn_local_data_landing"),
        )

        bcf.move(
            "demo_01_connections.yaml",
            destination=self.target_path / "demo_01_connections.yaml",
        )

        bcf_temp = conf.BaseConfFile(self.target_path)
        self.assertDictEqual(
            {
                "alias": "conn_local_data_landing",
                "endpoint": "file:///N/A/data/demo/landing",
                "type": "connection.LocalSystem",
            },
            bcf_temp.load(name="conn_local_data_landing"),
        )

        self.assertTrue(
            os.path.exists(self.target_path / "demo_01_connections.yaml")
        )

        if os.path.exists(self.target_path):
            shutil.rmtree(self.target_path)


@pytest.mark.usefixtures("test_path_to_cls")
class ConfigFileTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.demo_path: pathlib.Path = (
            self.test_path / "examples" / "conf" / "demo"
        )
        self.target_path: pathlib.Path = self.test_path / "conf_file_temp"

    def test_base_conf_read_file(self):
        cf = conf.ConfFile(self.demo_path)
        cf.move(
            path="demo_01_connections.yaml",
            destination=self.target_path / "demo_01_connections.yaml",
        )

        _stage_path = self.target_path / "demo_01_connections_stage.json"

        cf.create(path=_stage_path)
        self.assertTrue(os.path.exists(_stage_path))
        cf.save_stage(path=_stage_path, data=cf.load("conn_local_data_landing"))

        self.assertDictEqual(
            {
                "alias": "conn_local_data_landing",
                "endpoint": "file:///N/A/data/demo/landing",
                "type": "connection.LocalSystem",
            },
            cf.load_stage(path=_stage_path),
        )

        cf.save_stage(
            path=_stage_path,
            data={"temp_additional": cf.load("conn_local_data_landing")},
            merge=True,
        )

        cf.remove_stage(
            path=_stage_path,
            name="temp_additional",
        )

        self.assertDictEqual(
            {
                "alias": "conn_local_data_landing",
                "endpoint": "file:///N/A/data/demo/landing",
                "type": "connection.LocalSystem",
            },
            cf.load_stage(path=_stage_path),
        )

        if os.path.exists(self.target_path):
            shutil.rmtree(self.target_path)



import unittest
import warnings

import ddeutil.io.register as rgt
import pytest
from ddeutil.io.models import Params


@pytest.mark.usefixtures("test_path_to_cls")
class RegisterTestCase(unittest.TestCase):
    def setUp(self) -> None:
        warnings.simplefilter("ignore", category=ResourceWarning)
        self.param_config = Params.model_validate(
            {
                "engine": {
                    "paths": {
                        "conf": self.test_path / "examples/conf",
                        "data": self.root_path / "data",
                        "archive": self.root_path / "/data/.archive",
                    },
                    "flags": {"auto_update": True},
                },
                "stages": {
                    "raw": {"format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}"},
                    "persisted": {"format": "{naming:%s}.{version:v%m.%n.%c}"},
                },
            }
        )

    def test_register_init(self):
        register = rgt.Register(
            name="demo:conn_local_data_landing",
            config=self.param_config,
        )

        self.assertEqual("base", register.stage)
        self.assertDictEqual(
            {
                "alias": "conn_local_data_landing",
                "type": "connection.LocalSystem",
                "endpoint": "file:///N/A/data/demo/landing",
            },
            register.data(),
        )

        self.assertDictEqual(
            {
                "alias": "8568c1f93ae0441b5648ad7768c16c66",
                "type": "4a5a116820c0ae3f54269945c1d81863",
                "endpoint": "0e029e878279e73aba794c90cce8e48c",
            },
            register.data(hashing=True),
        )

        print("\nChange compare from metadata:", register.changed)

        rsg_raw = register.move(stage="raw")

        self.assertEqual("base", register.stage)
        self.assertEqual("raw", rsg_raw.stage)

        self.assertEqual(
            "8568c1f93ae0441b5648ad7768c16c66",
            rsg_raw.data(hashing=True)["alias"],
        )

        rgt.Register.reset(
            name="demo:conn_local_data_landing",
            config=self.param_config,
        )

    def test_register_without_config(self):
        with self.assertRaises(NotImplementedError) as context:
            rgt.Register(name="demo:conn_local_data_landing")
        self.assertEqual(
            (
                "This register instance can not do any actions because config "
                "param does not set."
            ),
            str(context.exception),
        )



import os.path
import pathlib
import shutil
import unittest
from typing import Dict

import ddeutil.io.config as conf
import pytest


@pytest.mark.usefixtures("test_path_to_cls")
class BaseConfigSQLiteTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.demo_path: pathlib.Path = (
            self.test_path / "examples" / "conf" / "demo"
        )
        self.target_path: pathlib.Path = self.test_path / "conf_sqlite_temp"

    def test_base_conf_read_file(self):
        _schemas: Dict[str, str] = {
            "name": "varchar(256) primary key",
            "shortname": "varchar(64) not null",
            "fullname": "varchar(256) not null",
            "data": "json not null",
            "updt": "datetime not null",
            "rtdt": "datetime not null",
            "author": "varchar(512) not null",
        }

        bc_sql = conf.ConfSQLite(self.target_path)
        bc_sql.create(table="demo.db/temp_table", schemas=_schemas)

        self.assertTrue(os.path.exists(self.target_path / "demo.db"))

        _data = {
            "conn_local_data_landing": {
                "name": "conn_local_data_landing",
                "shortname": "cldl",
                "fullname": "conn_local_data_landing",
                "data": {"first_row": {"key": "value"}},
                "updt": "2023-01-01 00:00:00",
                "rtdt": "2023-01-01 00:00:00",
                "author": "unknown",
            },
        }

        bc_sql.save_stage(table="demo.db/temp_table", data=_data)

        self.assertDictEqual(
            _data,
            bc_sql.load_stage(
                table="demo.db/temp_table",
            ),
        )
        if os.path.exists(self.target_path):
            shutil.rmtree(self.target_path)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import dataclasses
import logging
import sys
from functools import wraps
from typing import (
    Any,
    List,
    Optional,
    TypeVar,
    Union,
)
from urllib.parse import urlparse

from ddeutil.core import merge_dict
from ddeutil.core.dtutils import get_date

from .config import ConfAdapter, ConfFile, ConfSQLite

logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
logger = logging.getLogger(__name__)


ConfLoader = TypeVar("ConfLoader", bound="BaseConfLoader")


class BaseConfLoader:
    """Base Config Loader Object for get data from any config file or
    table with name.
    """

    conf_prefix: str = ""

    conf_file_extension: str = ""

    conf_file_initial: Any = {}

    conf_sqlite_schema: dict = {}

    def __init__(
        self,
        endpoint: Optional[str],
        name: str,
        *,
        environ: Optional[str] = None,
    ):
        """Initialization of base metadata object. The metadata can implement
        with a json file or a table in any database like SQLite.

        :param endpoint: str

        :param name: str

        :param environ: Optional[str]
        """
        _environ: str = f".{_env}" if (_env := (environ or "")) else ""
        # TODO: Case endpoint is None
        url = urlparse(endpoint)
        self._cf_loader_type = url.scheme
        self._cf_loader_endpoint = f"{url.netloc}/{url.path}"
        self._cf_name: str = name

        if self._cf_loader_type == "file":
            # Case: file : the data in file does not have schemas.
            self._cf_filename: str = (
                f"{self.conf_prefix}{_environ}.{self.conf_file_extension}"
            )
            self.loading: ConfAdapter = ConfFile(
                path=self._cf_loader_endpoint, _type=self.conf_file_extension
            )
            self.loading.create(
                name=self._cf_filename,
                initial_data=self.conf_file_initial,
            )
        elif self._cf_loader_type == "sqlite":
            # Case: sqlite : the data must use `conf_sqlite_schema`
            # for table creation.
            self._cf_filename: str = (
                f"{self.conf_prefix}{_environ}.db/tbl_{self.conf_prefix}"
            )
            self.loading: ConfAdapter = ConfSQLite(self._cf_loader_endpoint)
            self.loading.create(
                name=self._cf_filename, schemas=self.conf_sqlite_schema
            )
        else:
            raise NotImplementedError(
                f"{self.__class__.__name__} does not support "
                f"type: {self._cf_loader_type!r}."
            )

    def load(self) -> dict:
        """Return data from filename."""
        return self.loading.load_stage(name=self._cf_filename).get(
            self._cf_name, {}
        )

    def save(self, data: Union[dict, list]) -> None:
        """Saving data to source from filename."""
        self.loading.save_stage(name=self._cf_filename, data=data, merge=True)

    def remove(self) -> None:
        """Remove data from filename."""
        self.loading.remove_stage(
            name=self._cf_filename,
            data_name=self._cf_name,
        )


class ConfMetadata(BaseConfLoader):
    """Config Metadata Object for get data from metadata file or table"""

    conf_prefix: str = "metadata"

    conf_file_extension: str = "json"

    conf_file_initial: dict = {}

    conf_sqlite_schema: dict = {
        "name": "varchar(256) primary key",
        "shortname": "varchar(64) not null",
        "fullname": "varchar(256) not null",
        "data": "json not null",
        "updt": "datetime not null",
        "rtdt": "datetime not null",
        "author": "varchar(512) not null",
    }

    def save(self, data: dict) -> None:
        """Saving data to source with name mapping from filename."""
        self.loading.save_stage(
            name=self._cf_filename,
            data={self._cf_name: data},
            merge=True,
        )


@dataclasses.dataclass()
class Message:
    """Message Data Object for get message string and concat with `||` when
    pull result.
    """

    _message: str
    _messages: List[str] = dataclasses.field(default_factory=list)

    @property
    def messages(self) -> str:
        return self._message

    @messages.setter
    def messages(self, msg: str):
        self._messages.append(msg)
        self._message += f"||{msg}" if self._message else f"{msg}"


Logging = TypeVar("Logging", bound="ConfLogging")


def saving(func):
    @wraps(func)
    def wraps_saving(*args, **kwargs):
        self: Logging = args[0]
        _level: str = func.__name__.split("_")[-1].upper()
        if (
            (msg := args[1])
            and (getattr(logging, _level) >= self._cf_logger.level)
            and (self._cf_auto_save or kwargs.get("force", False))
        ):
            self.save(
                data=self.setup(
                    {
                        # TODO: converter msg string before save.
                        "message": msg,
                        "status": _level,
                    }
                )
            )
        return func(*args, **kwargs)

    return wraps_saving


class ConfLogging(BaseConfLoader):
    """Config Logging Object for log message from any change from register
    or loader process.
    """

    conf_prefix: str = "logging"

    conf_file_extension: str = "csv"

    conf_file_initial: list = []

    conf_sqlite_schema: dict = {
        "parent_hash_id": "varchar(64) not null",
        "hash_id": "varchar(64) primary key not null",
        "conf_name": "varchar(256) not null",
        "message": "text",
        "updt": "datetime not null",
        "status": "varchar(64) not null",
        "author": "varchar(512) not null",
    }

    def __init__(
        self,
        endpoint: str,
        name: str,
        _logger: logging.Logger,
        *,
        environ: Optional[str] = None,
        setup: Optional[dict] = None,
        auto_save: bool = False,
    ):
        super().__init__(endpoint, name, environ=environ)
        self._cf_logger: logging.Logger = _logger
        self._cf_msgs: List[Optional[dict]] = []
        self._cf_parent_hash: str = str(int(get_date("datetime").timestamp()))
        self._cf_setup: dict = merge_dict(
            {"conf_name": self._cf_name}, (setup or {})
        )
        self._cf_auto_save: bool = auto_save

    def setup(self, data) -> dict:
        _now = get_date("datetime")
        return merge_dict(
            self._cf_setup,
            {
                "parent_hash_id": self._cf_parent_hash,
                "hash_id": str(int(_now.timestamp())),
                "updt": _now.strftime("%Y-%m-%d %H:%M:%S"),
            },
            data,
        )

    def save_logging(self):
        if self.is_pulled:
            self.save(data=self._cf_msgs)
            self._cf_msgs: List[Optional[dict]] = []

    def debug(self, msg):
        self._cf_logger.debug(msg)

    def info(self, msg):
        self._cf_logger.info(msg)

    def warning(self, msg):
        self._cf_logger.warning(msg)

    def critical(self, msg):
        self._cf_logger.critical(msg)

    @saving
    def p_debug(self, msg, force: bool = False):
        if (
            (logging.DEBUG >= self.level)
            and (not self._cf_auto_save)
            and (not force)
        ):
            self._cf_msgs.append(
                self.setup({"message": msg, "status": "DEBUG"}),
            )
        self._cf_logger.debug(msg)

    @saving
    def p_info(self, msg, force: bool = False):
        if (
            logging.INFO >= self.level
            and (not self._cf_auto_save)
            and (not force)
        ):
            self._cf_msgs.append(self.setup({"message": msg, "status": "INFO"}))
        self._cf_logger.info(msg)

    @saving
    def p_warning(self, msg, force: bool = False):
        if (
            logging.WARNING >= self.level
            and (not self._cf_auto_save)
            and (not force)
        ):
            self._cf_msgs.append(
                self.setup({"message": msg, "status": "WARNING"}),
            )
        self._cf_logger.warning(msg)

    @saving
    def p_critical(self, msg, force: bool = False):
        if (
            logging.CRITICAL >= self.level
            and (not self._cf_auto_save)
            and (not force)
        ):
            self._cf_msgs.append(
                self.setup({"message": msg, "status": "CRITICAL"}),
            )
        self._cf_logger.critical(msg)

    @property
    def is_pulled(self) -> bool:
        return len(self._cf_msgs) > 0

    @property
    def level(self) -> int:
        return self._cf_logger.level


def test_meta_sqlite():
    _meta = ConfMetadata(
        "sqlite://D:/korawica/Work/dev02_miniproj/GITHUB/dde-object-defined/data",
        # 'file://D:/korawica/Work/dev02_miniproj/GITHUB/dde-object-defined/data',
        "conn_local_file2",
    )
    _meta.save(
        data={
            "name": "conn_local_file2",
            "shortname": "t3",
            "fullname": "super_test",
            "data": {"data": "ASFGSDFE13123"},
            "updt": "2022-01-02 00:00:00",
            "rtdt": "2022-01-01 00:00:00",
            "author": "unknown",
        }
    )
    print(_meta.load())


def test_logging_file():
    _log = ConfLogging(
        "file://D:/korawica/Work/dev02_miniproj/GITHUB/dde-object-defined/data/logs",
        "conn_local_file",
        _logger=logger,
        auto_save=False,
    )
    _log.p_info("test log data from info level", force=True)
    _log.p_debug("test log data from debug level1")
    logger.setLevel(logging.INFO)
    _log.p_debug("test log data from debug level2")
    print(_log.is_pulled)


if __name__ == "__main__":
    test_meta_sqlite()
    # test_logging_file()



import os
from typing import (
    Callable,
    Dict,
    Optional,
)

from .settings import SettingRegex


def add_newline(text: str, newline: Optional[str] = None) -> str:
    nl: str = newline or "\n"
    return f"{text}{nl}" if not text.endswith(nl) else text


def search_env_replace(
    contents: str,
    *,
    raise_if_default_not_exists: bool = False,
    default_value: str = "N/A",
    escape_replaced: str = "ESC",
    caller: Callable[[str], str] = (lambda x: x),
) -> str:
    """Prepare content data before parse to any file loading method"""
    shifting: int = 0
    replaces: dict = {}
    replaces_esc: dict = {}
    for content in SettingRegex.RE_ENV_SEARCH.finditer(contents):
        search: str = content.group(1)
        if not (escaped := content.group("escaped")):
            variable: str = content.group("braced")
            default: str = content.group("braced_default")
            if not default and raise_if_default_not_exists:
                raise ValueError(
                    f"Could not find default value for {variable} "
                    f"in `.yaml` file"
                )
            elif not variable:
                raise ValueError(
                    f"Value {search!r} in `.yaml` file has something wrong "
                    f"with regular expression"
                )
            replaces[search] = caller(
                os.environ.get(variable, default) or default_value
            )
        elif "$" in escaped:
            span = content.span()
            search = f"${{{escape_replaced}{escaped}}}"
            contents = (
                contents[: (span[0] + shifting)]
                + search
                + contents[(span[1] + shifting) :]
            )
            shifting += len(search) - (span[1] - span[0])
            replaces_esc[search] = "$"
    for _replace in sorted(replaces, reverse=True):
        contents = contents.replace(_replace, replaces[_replace])
    for _replace in sorted(replaces_esc, reverse=True):
        contents = contents.replace(_replace, replaces_esc[_replace])
    return contents


def search_env(
    contents: str,
    *,
    keep_newline: bool = False,
    default: Optional[str] = None,
) -> Dict[str, str]:
    """Prepare content data from .env string format before load
    to the OS environment.

    :ref:
        - python-dotenv
            ref: https://github.com/theskumar/python-dotenv
    """
    _default: str = default or ""
    env: Dict[str, str] = {}
    for content in SettingRegex.RE_DOTENV.finditer(contents):
        name: str = content.group("name")

        # Remove leading/trailing whitespace
        _value: str = (content.group("value") or "").strip()

        if not _value:
            raise ValueError(
                f"Value {name:!r} in `.env` file does not set value "
                f"of variable"
            )
        value: str = _value if keep_newline else "".join(_value.splitlines())
        quoted: Optional[str] = None

        # Remove surrounding quotes
        if m2 := SettingRegex.RE_ENV_VALUE_QUOTED.match(value):
            quoted: str = m2.group("quoted")
            value: str = m2.group("value")

        if quoted == "'":
            env[name] = value
            continue
        elif quoted == '"':
            # Unescape all chars except $ so variables
            # can be escaped properly
            value: str = SettingRegex.RE_ENV_ESCAPE.sub(r"\1", value)

        # Substitute variables in a value
        env[name] = __search_var(value, env, default)
    return env


def __search_var(
    value: str,
    env: Dict[str, str],
    default: Optional[str] = None,
) -> str:
    _default: str = default or ""
    for sub_content in SettingRegex.RE_DOTENV_VAR.findall(value):
        replace: str = "".join(sub_content[1:-1])
        if sub_content[0] != "\\":
            # Replace it with the value from the environment
            replace: str = env.get(
                sub_content[-1],
                os.environ.get(sub_content[-1], _default),
            )
        value: str = value.replace("".join(sub_content[:-1]), replace)
    return value



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import abc
import contextlib
import inspect
import json
import logging
import os
import pathlib
import shutil
import sqlite3
import sys
from datetime import datetime
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    NoReturn,
    Optional,
    Type,
    Union,
)

from ddeutil.core import merge_dict

from .base import (
    Json,
    OpenFile,
    YamlEnv,
)
from .base.pathutils import PathSearch, remove_file
from .exceptions import ConfigArgumentError

logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
logger = logging.getLogger(__name__)


class BaseConfFile:
    """Base Config File object for getting data with `.yaml` format and mapping
    environment variables to the content data.
    """

    def __init__(
        self,
        path: Union[str, pathlib.Path],
        *,
        compress: Optional[str] = None,
        auto_create: bool = True,
        open_fil: Optional[Type[OpenFile]] = None,
        fil_fmt_exc: Optional[List[str]] = None,
    ):
        self.path: pathlib.Path = (
            pathlib.Path(path) if isinstance(path, str) else path
        )
        self.compress = compress
        if not os.path.exists(self.path):
            if not auto_create:
                raise FileNotFoundError(f"Path {path} does not exists.")
            os.makedirs(self.path)
        self.open_fil: Type[OpenFile] = open_fil or YamlEnv
        self.fil_fmt_exc: List[str] = fil_fmt_exc or [".json", ".toml"]

    def load(
        self,
        name: str,
        *,
        order: int = 1,
    ) -> Dict[Any, Any]:
        """Return configuration data from name of the config.

        :param name: A name of config key that want to search in the path.
        :type name: str
        :param order: An order number that want to get from ordered list
            of duplicate data.
        :type order: int (Default 1)
        """
        rs: List[Dict[Any, Any]]
        if rs := [
            merge_dict({"alias": name}, data)
            for file in self.files(excluded=self.fil_fmt_exc)
            if (
                data := self.open_fil(path=file, compress=self.compress)
                .read()
                .get(name)
            )
        ]:
            try:
                if order > len(rs):
                    raise IndexError
                return sorted(
                    rs,
                    key=lambda x: (
                        datetime.fromisoformat(x.get("version", "1990-01-01")),
                        len(x),
                    ),
                    reverse=False,
                )[-order]
            except IndexError:
                logging.warning(
                    f"Does not load config data with order: -{order} "
                    f"and name: {name!r}."
                )
        return {}

    def files(
        self,
        path: Optional[str] = None,
        name: Optional[str] = None,
        *,
        excluded: Optional[list] = None,
    ) -> Iterator:
        """Return all files that exists in the loading path."""
        return filter(
            lambda x: os.path.isfile(x),
            PathSearch.from_dict(
                {
                    "root": (path or str(self.path.resolve())),
                    "exclude_name": excluded,
                }
            ).pick(filename=(name or "*")),
        )

    def move(
        self,
        path: Union[str, pathlib.Path],
        destination: Union[str, pathlib.Path],
        *,
        auto_create: bool = True,
    ):
        """Copy filename to destination path."""
        if auto_create:
            os.makedirs(os.path.dirname(destination), exist_ok=True)
        shutil.copy(self.path / path, destination)


class ConfAdapter(abc.ABC):
    """Config Adapter abstract object."""

    @abc.abstractmethod
    def load_stage(self, name: str) -> dict:
        raise NotImplementedError

    @abc.abstractmethod
    def save_stage(
        self,
        name: str,
        data: dict,
        merge: bool = False,
    ) -> NoReturn:
        raise NotImplementedError

    @abc.abstractmethod
    def remove_stage(
        self,
        name: str,
        data_name: str,
    ) -> NoReturn:
        raise NotImplementedError

    @abc.abstractmethod
    def create(self, name: str, **kwargs) -> NoReturn:
        raise NotImplementedError


class ConfFile(BaseConfFile, ConfAdapter):
    """Config File Loading Object for get data from configuration and stage."""

    def __init__(
        self,
        path: Union[str, pathlib.Path],
        *,
        compress: Optional[str] = None,
        auto_create: bool = True,
        open_fil: Optional[Type[OpenFile]] = None,
        fil_fmt_exc: Optional[List[str]] = None,
        open_fil_stg: Optional[Type[OpenFile]] = None,
    ):
        """Main initialize of config file loading object.

        :param path: A path of files to action.
        :type path: Union[str, pathlib.Path]
        :param compress: Optional[str] : A compress type of action file.
        """
        super().__init__(
            path,
            compress=compress,
            auto_create=auto_create,
            open_fil=open_fil,
            fil_fmt_exc=fil_fmt_exc,
        )
        self.open_fil_stg: Type[OpenFile] = open_fil_stg or Json

    def load_stage(
        self,
        path: Union[str, pathlib.Path],
        default: Optional[Any] = None,
    ) -> Union[Dict[Any, Any], List[Any]]:
        """Return content data from file with filename, default empty dict."""
        try:
            return self.open_fil_stg(
                path=path,
                compress=self.compress,
            ).read()
        except FileNotFoundError:
            return default if (default is not None) else {}

    def save_stage(
        self,
        path: Union[str, pathlib.Path],
        data: Union[Dict[Any, Any], List[Any]],
        *,
        merge: bool = False,
    ) -> NoReturn:
        """Write content data to file with filename. If merge is true, it will
        load current data from file and merge the data content together
        before write.
        """
        if not merge:
            self.open_fil_stg(path, compress=self.compress).write(data)
            return
        elif merge and (
            "mode"
            in inspect.getfullargspec(self.open_fil_stg.write).annotations
        ):
            self.open_fil_stg(path, compress=self.compress).write(
                **{
                    "data": data,
                    "mode": "a",
                }
            )
            return

        all_data: Union[dict, list] = self.load_stage(path=path)
        try:
            if isinstance(all_data, list):
                _merge_data: Union[dict, list] = all_data
                if isinstance(data, dict):
                    _merge_data.append(data)
                else:
                    _merge_data.extend(data)
            else:
                _merge_data: Union[dict, list] = merge_dict(all_data, data)
            self.open_fil_stg(path, compress=self.compress).write(_merge_data)
        except TypeError as err:
            remove_file(path=path)
            if all_data:
                self.open_fil_stg(path, compress=self.compress).write(
                    all_data,
                )
            raise err

    def remove_stage(self, path: str, name: str) -> NoReturn:
        """Remove data by name from file with filename."""
        if all_data := self.load_stage(path=path):
            all_data.pop(name, None)
            self.open_fil_stg(path, compress=self.compress).write(
                all_data,
            )

    def create(
        self,
        path: Union[str, pathlib.Path],
        *,
        initial_data: Optional[Any] = None,
    ) -> NoReturn:
        """Create filename in path."""
        if not os.path.exists(path):
            self.save_stage(
                path=path,
                data=({} if initial_data is None else initial_data),
                merge=False,
            )


class BaseConfSQLite:
    """Base Config SQLite object for getting data with SQLite database from
    file storage."""

    def __init__(
        self,
        path: Union[str, pathlib.Path],
        *,
        auto_create: bool = True,
    ):
        self.path: pathlib.Path = (
            pathlib.Path(path) if isinstance(path, str) else path
        )
        if not os.path.exists(self.path):
            if not auto_create:
                raise FileNotFoundError(f"Path {path} does not exists.")
            os.makedirs(self.path, exist_ok=True)

    @contextlib.contextmanager
    def connect(self, database: str):
        """Return SQLite Connection context."""
        _conn = sqlite3.connect(self.path / database, timeout=3)
        _conn.row_factory = self.dict_factory
        try:
            yield _conn
        except sqlite3.Error as err:
            logger.error(err)
            raise ConfigArgumentError(
                "syntax", f"SQLite syntax error {err}"
            ) from err
        _conn.commit()
        _conn.close()

    @staticmethod
    def dict_factory(cursor, row):
        """Result of dictionary factory.

        :note:
            Another logic of dict factory.

            - dict(
                [
                    (col[0], row[idx])
                    for idx, col in enumerate(cursor.description)
                ]
            )

            - dict(zip([col[0] for col in cursor.description], row))
        """
        return {col[0]: row[idx] for idx, col in enumerate(cursor.description)}


class ConfSQLite(BaseConfSQLite, ConfAdapter):
    def load_stage(
        self,
        table: str,
        default: Optional[Dict[Any, Any]] = None,
    ) -> Dict[Any, Any]:
        """Return content data from database with table name, default empty
        dict."""
        _db, _table = table.rsplit("/", maxsplit=1)
        with self.connect(_db) as conn:
            cur = conn.cursor()
            cur.execute(f"select * from {_table};")
            if result := cur.fetchall():
                return {_["name"]: self.convert_type(_) for _ in result}
            return default if (default is not None) else {}

    def save_stage(
        self,
        table: str,
        data: dict,
        merge: bool = False,
    ) -> NoReturn:
        """Write content data to database with table name. If merge is true, it
        will update or insert the data content.
        """
        _db, _table = table.rsplit("/", maxsplit=1)
        _data: dict = self.prepare_values(data.get(list(data.keys())[0]))
        with self.connect(_db) as conn:
            cur = conn.cursor()
            doing: str = "nothing"
            if merge:
                _doing_list = [
                    f"{_} = excluded.{_}" for _ in _data if _ != "name"
                ]
                doing: str = f'update set {", ".join(_doing_list)}'
            query: str = (
                f'insert into {_table} ({", ".join(_data.keys())}) values '
                f'({":" + ", :".join(_data.keys())}) '
                f"on conflict ( name ) do {doing};"
            )
            cur.execute(query, _data)

    def remove_stage(
        self,
        table: str,
        data_name: str,
    ) -> NoReturn:
        """Remove data by name from table in database with table name."""
        _db, _table = table.rsplit("/", maxsplit=1)
        with self.connect(_db) as conn:
            cur = conn.cursor()
            query: str = f"delete from {_table} where name = '{data_name}';"
            cur.execute(query)

    def create(
        self,
        table: str,
        schemas: Optional[dict] = None,
    ) -> NoReturn:
        """Create table in database."""
        if not schemas:
            raise ConfigArgumentError(
                "schemas",
                (
                    f"in `create` method of {self.__class__.__name__} "
                    f"should have value."
                ),
            )
        _schemas: str = ", ".join([f"{k} {v}" for k, v in schemas.items()])
        _db, _table = table.rsplit("/", maxsplit=1)
        with self.connect(_db) as conn:
            cur = conn.cursor()
            cur.execute(f"create table if not exists {_table} ({_schemas})")

    @staticmethod
    def prepare_values(
        values: dict,
    ) -> Dict[str, Union[str, int, float]]:
        """Return prepare value with dictionary type to string
        to source system.
        """
        results: dict = values.copy()
        for _ in values:
            if isinstance(values[_], dict):
                results[_] = json.dumps(values[_])
        return results

    @staticmethod
    def convert_type(
        data: dict,
        key: Optional[str] = None,
    ) -> dict:
        """Return converted value from string to dictionary
        from source system.
        """
        _key: str = key or "data"
        _results: dict = data.copy()
        _results[_key] = json.loads(data[_key])
        return _results


__all__ = (
    "ConfAdapter",
    "ConfFile",
    "ConfSQLite",
    "OpenFile",
)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import datetime
import pathlib
import re
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from ddeutil.core import merge_dict

# from .base import YamlEnv
# from .exceptions import ConfigArgumentError
from ddeutil.io.base import YamlEnv
from ddeutil.io.exceptions import ConfigArgumentError
from pydantic import (
    BaseModel,
    Field,
    FieldValidationInfo,
    field_validator,
)

FMT_NAMES: Tuple[str, ...] = (
    "naming",
    "domain",
    "environ",
    "timestamp",
    "version",
    "compress",
    "extension",
)

RULE_FIX: Tuple[str, ...] = (
    "timestamp",
    "version",
    "compress",
)


class RuleData(BaseModel):
    """Rule Data Model
    .. example::
        {
            "timestamp": {
                "minutes": 15
            },
            "excluded": [],
            "compress": None,
        }
    """

    timestamp: Optional[Dict[str, int]] = Field(default_factory=dict)

    version: Optional[str] = Field(default=None)
    excluded: Optional[List[str]] = Field(default_factory=list)
    compress: Optional[str] = Field(default=None)


class StageData(BaseModel):
    """
    .. example::
        {
            "raw": {
                "format": "",
                rules: {},
            },
        }
    """

    alias: str
    rules: RuleData = Field(default_factory=RuleData)
    format: str
    layer: int = Field(default=0)

    @field_validator("format", mode="before")
    @classmethod
    def validate_format(cls, value, info: FieldValidationInfo):
        # Validate the name in format string should contain any format name.
        if not (
            _searches := re.findall(
                r"{(?P<name>\w+):?(?P<format>[^{}]+)?}",
                value,
            )
        ):
            raise ConfigArgumentError(
                "format",
                (
                    f'This `{info.data["alias"]}` stage format dose not '
                    f"include any format name, the stage file was duplicated."
                ),
            )

        # Validate the name in format string should exist in `FMT_NAMES`.
        if any((_search[0] not in FMT_NAMES) for _search in _searches):
            raise ConfigArgumentError(
                "format",
                "This stage have an unsupported format name.",
            )
        return value

    @field_validator("format", mode="after")
    @classmethod
    def validate_rule_relate_with_format(cls, value, info: FieldValidationInfo):
        # Validate a format of stage that relate with rules.
        for validator in RULE_FIX:
            if getattr(info.data.get("rules", {}), validator, None) and (
                validator not in value
            ):
                raise ConfigArgumentError(
                    (
                        "format",
                        validator,
                    ),
                    (
                        f"This stage set `{validator}` rule but does not have "
                        f"a `{validator}` format name in the format."
                    ),
                )
        return value


class PathData(BaseModel):
    root: Union[str, pathlib.Path] = Field(default_factory=pathlib.Path)
    data: pathlib.Path = Field(default="data", validate_default=True)
    conf: pathlib.Path = Field(default="conf", validate_default=True)
    archive: pathlib.Path = Field(default=".archive", validate_default=True)

    @field_validator("root", mode="before")
    @classmethod
    def prepare_root(cls, v):
        if isinstance(v, str):
            return pathlib.Path(v)
        return v

    @field_validator("data", "conf", "archive", mode="before")
    @classmethod
    def prepare_path_from_str(
        cls,
        v,
        info: FieldValidationInfo,
    ) -> pathlib.Path:
        _root: pathlib.Path = info.data["root"]
        return v if isinstance(v, pathlib.Path) else (_root / v)


class FlagData(BaseModel):
    archive: bool = Field(default=False)
    auto_update: bool = Field(default=False)


class ValueData(BaseModel):
    datetime_fmt: str = Field(default="%Y-%m-%d %H:%M:%S")
    excluded_keys: Tuple[str, ...] = Field(
        default=(
            "version",
            "updt",
        )
    )


class EngineData(BaseModel):
    paths: PathData = Field(default_factory=PathData)
    flags: FlagData = Field(default_factory=FlagData)
    values: ValueData = Field(default_factory=ValueData)


class Params(BaseModel, validate_assignment=True):
    stages: Dict[str, StageData] = Field(default_factory=dict)
    engine: EngineData = Field(default_factory=EngineData)

    @classmethod
    def from_file(cls, path: Union[str, pathlib.Path]):
        cls._origin_path = path
        return cls.model_validate(YamlEnv(path).read())

    @field_validator("stages", mode="before")
    @classmethod
    def order_layer(cls, value: Dict[str, Dict[Any, Any]]):
        for i, k in enumerate(value, start=1):
            value[k] = merge_dict(
                value[k].copy(),
                {
                    "layer": i,
                    "alias": k,
                },
            )
        return value

    @property
    def stage_final(self) -> str:
        return max(self.stages.items(), key=lambda i: i[1].layer)[0]

    @property
    def stage_first(self) -> str:
        return min(self.stages.items(), key=lambda i: i[1].layer)[0]

    def get_stage(self, name: str):
        if name == "base":
            return StageData.model_validate(
                {
                    "format": "{naming}_{timestamp}",
                    "layer": 0,
                    "alias": "base",
                }
            )
        elif name not in self.stages:
            raise ConfigArgumentError(
                "stage",
                (
                    f"Cannot get stage: {name!r} because it does not set "
                    f"in `parameters.yaml`"
                ),
            )
        return self.stages[name].model_copy()

    def refresh(self, path: Optional[str] = None) -> Params:
        _origin_path = path or self._origin_path
        self.__dict__.update(self.from_file(path=_origin_path).__dict__)
        return self


def gt(self: StageData, other: StageData):
    return self.layer > other.layer


class RegisterConfData(BaseModel):
    author: str = Field(default="unknown")
    auto_update: bool = Field(default=True)
    force_exists: bool = Field(default=False)


class RegisterMetaData(BaseModel, validate_assignment=True):
    name: str
    shortname: str
    fullname: str
    data: Dict[str, Any]
    updt: datetime.datetime
    rtdt: datetime.datetime
    author: str


if __name__ == "__main__":
    print((pathlib.Path(__file__).parent / "../../..").resolve())
    print((pathlib.Path(__file__).parent / "../../.." / "").resolve())



import pathlib
import unittest

import ddeutil.io.models as md


class ModelTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.maxDiff = None

    def test_model_path_data(self):
        p = md.PathData.model_validate(
            {
                "data": pathlib.Path("."),
                "conf": pathlib.Path("."),
                "archive": pathlib.Path("."),
            }
        )

        self.assertDictEqual(
            {
                "data": pathlib.Path("."),
                "conf": pathlib.Path("."),
                "archive": pathlib.Path("."),
                "root": pathlib.Path("."),
            },
            p.model_dump(),
        )

    def test_model_path_data_with_root(self):
        p = md.PathData.model_validate(
            {
                "root": "./src/",
            }
        )

        self.assertDictEqual(
            {
                "data": pathlib.Path("./src/data"),
                "conf": pathlib.Path("./src/conf"),
                "archive": pathlib.Path("./src/.archive"),
                "root": pathlib.Path("./src/"),
            },
            p.model_dump(),
        )

    def test_model_rule_data(self):
        rule = md.RuleData.model_validate({})

        self.assertDictEqual(
            {
                "timestamp": {},
                "version": None,
                "excluded": [],
                "compress": None,
            },
            rule.model_dump(),
        )

    def test_model_stage_data(self):
        stage = md.StageData.model_validate(
            {
                "alias": "persisted",
                "format": "{timestamp:%Y-%m-%d}{naming:%c}.json",
                "rules": {
                    "timestamp": {"minutes": 15},
                },
            }
        )

        self.assertDictEqual(
            {
                "alias": "persisted",
                "format": "{timestamp:%Y-%m-%d}{naming:%c}.json",
                "rules": {
                    "timestamp": {"minutes": 15},
                    "version": None,
                    "excluded": [],
                    "compress": None,
                },
                "layer": 0,
            },
            stage.model_dump(),
        )

    def test_model_params(self):
        params = md.Params.model_validate(
            {
                "stages": {
                    "raw": {"format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}"},
                    "persisted": {"format": "{naming:%s}.{version:v%m.%n.%c}"},
                    "curated": {
                        "format": "{domain:%s}_{naming:%s}.{compress:%-g}"
                    },
                }
            }
        )
        self.assertDictEqual(
            {
                "stages": {
                    "raw": {
                        "format": "{naming:%s}.{timestamp:%Y%m%d_%H%M%S}",
                        "alias": "raw",
                        "rules": {
                            "timestamp": {},
                            "version": None,
                            "excluded": [],
                            "compress": None,
                        },
                        "layer": 1,
                    },
                    "persisted": {
                        "format": "{naming:%s}.{version:v%m.%n.%c}",
                        "alias": "persisted",
                        "rules": {
                            "timestamp": {},
                            "version": None,
                            "excluded": [],
                            "compress": None,
                        },
                        "layer": 2,
                    },
                    "curated": {
                        "format": "{domain:%s}_{naming:%s}.{compress:%-g}",
                        "alias": "curated",
                        "rules": {
                            "timestamp": {},
                            "version": None,
                            "excluded": [],
                            "compress": None,
                        },
                        "layer": 3,
                    },
                },
                "engine": {
                    "values": {
                        "datetime_fmt": "%Y-%m-%d %H:%M:%S",
                        "excluded_keys": ("version", "updt"),
                    },
                    "flags": {
                        "archive": False,
                        "auto_update": False,
                    },
                    "paths": {
                        "archive": pathlib.Path(".archive"),
                        "conf": pathlib.Path("conf"),
                        "data": pathlib.Path("data"),
                        "root": pathlib.Path("."),
                    },
                },
            },
            params.model_dump(),
        )
